;;; -*- Mode: LISP; Syntax: COMMON-LISP; Package: HUNCHENTOOT; Base: 10 -*-
;;; $Header: /usr/local/cvsrep/hunchentoot/misc.lisp,v 1.6 2007/01/01 23:50:30 edi Exp $

;;; Copyright (c) 2004-2007, Dr. Edmund Weitz. All rights reserved.

;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:

;;;   * Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.

;;;   * Redistributions in binary form must reproduce the above
;;;     copyright notice, this list of conditions and the following
;;;     disclaimer in the documentation and/or other materials
;;;     provided with the distribution.

;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
;;; DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
;;; GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(in-package :hunchentoot)

(let ((scanner-hash (make-hash-table :test #'equal)))
  (defun scanner-for-get-param (param-name)
    "Returns a CL-PPCRE scanner which matches a GET parameter in a
URL.  Scanners are memoized in SCANNER-HASH once they are created."
    (or (gethash param-name scanner-hash)
        (setf (gethash param-name scanner-hash)
                (create-scanner
                 `(:alternation
                   ;; session=value at end of URL
                   (:sequence
                    (:char-class #\? #\&)
                    ,param-name
                    #\=
                    (:greedy-repetition 0 nil (:inverted-char-class #\&))
                    :end-anchor)
                   ;; session=value with other parameters following
                   (:sequence
                    (:register (:char-class #\? #\&))
                    ,param-name
                    #\=
                    (:greedy-repetition 0 nil (:inverted-char-class #\&))
                    #\&))))))
  (defun add-cookie-value-to-url (url &key (cookie-name *session-cookie-name*)
                                           (value (session-cookie-value))
                                           (replace-ampersands-p t))
    "Removes all GET parameters named COOKIE-NAME from URL and then
adds a new GET parameter with the name COOKIE-NAME and the value
VALUE.  If REPLACE-AMPERSANDS-P is true all literal ampersands in URL
are replaced with '&amp;'. The resulting URL is returned."
    (unless url
      ;; see URL-REWRITE:*URL-REWRITE-FILL-TAGS*
      (setq url (request-uri *request*)))
    (setq url (regex-replace-all (scanner-for-get-param cookie-name) url "\\1"))
    (when value
      (setq url (format nil "~A~:[?~;&~]~A=~A"
                        url 
                        (find #\? url)
                        cookie-name
                        (url-encode value))))
    (when replace-ampersands-p
      (setq url (regex-replace-all "&" url "&amp;")))
    url))

(defun maybe-rewrite-urls-for-session (html &key (cookie-name *session-cookie-name*)
                                                 (value (session-cookie-value)))
  "Rewrites the HTML page HTML such that the name/value pair
COOKIE-NAME/COOKIE-VALUE is inserted if the client hasn't sent a
cookie of the same name but only if *REWRITE-FOR-SESSION-URLS* is
true.  See the docs for URL-REWRITE:REWRITE-URLS."
  (cond ((or (not *rewrite-for-session-urls*)
             (null value)
             (cookie-in cookie-name))
          html)
        (t
          (with-input-from-string (*standard-input* html)
            (with-output-to-string (*standard-output*)
              (url-rewrite:rewrite-urls
               (lambda (url)
                 (add-cookie-value-to-url url
                                          :cookie-name cookie-name
                                          :value value))))))))

(defmethod dispatch-request (dispatch-table)
  "Dispatches *REQUEST* based upon rules in the DISPATCH-TABLE.
This method provides the default Hunchentoot behavior."
  (loop for dispatcher in dispatch-table
        for action = (funcall dispatcher *request*)
        when action return (funcall action)
        finally (setf (return-code *reply*) +http-not-found+)))

(defun default-dispatcher (request)
  "Default dispatch function which handles every request with the
function stored in *DEFAULT-HANDLER*."
  (declare (ignore request))
  *default-handler*)

(defun default-handler ()
  "The handler that is supposed to serve the request if no other
handler is called."
  (log-message :info "Default handler called for script ~A" (script-name))
  (format nil "<html><head><title>Hunchentoot</title></head><body><h2>Hunchentoot Default Page</h2><p>This the Hunchentoot default page. You're most likely seeing it because the server administrator hasn't set up a custom default page yet.</p><p>Hunchentoot is a web server written in <a href='http://www.lisp.org/'>Common Lisp</a>.  More info about Hunchentoot can be found at <a href='http://weitz.de/hunchentoot/'>http://weitz.de/hunchentoot/</a>.</p></p><p><hr>~A</p></body></html>"
          (address-string)))

(defun create-prefix-dispatcher (prefix page-function)
  "Creates a dispatch function which will dispatch to the
function denoted by PAGE-FUNCTION if the file name of the current
request starts with the string PREFIX."
  (lambda (request)
    (let ((mismatch (mismatch (script-name request) prefix
                              :test #'char=)))
      (and (or (null mismatch)
               (>= mismatch (length prefix)))
           page-function))))

(defun create-regex-dispatcher (regex page-function)
  "Creates a dispatch function which will dispatch to the
function denoted by PAGE-FUNCTION if the file name of the current
request matches the CL-PPCRE regular expression REGEX."
  (let ((scanner (create-scanner regex)))
    (lambda (request)
      (and (scan scanner (script-name request))
           page-function))))

(defun handle-static-file (path &optional content-type)
  "A function which acts like a Hunchentoot handler for the file
denoted by PATH.  Send a content type header corresponding to
CONTENT-TYPE or \(if that is NIL) tries to determine the content
type via the file's suffix."
  (unless (and path
               (or (pathname-name path)
                   (pathname-type path)))
    ;; not a file
    (setf (return-code) +http-bad-request+)
    (throw 'handler-done nil))
  (unless (probe-file path)
    ;; does not exist
    (setf (return-code) +http-not-found+)
    (throw 'handler-done nil))
  (let ((time (or (file-write-date path) (get-universal-time))))
    (setf (content-type) (or content-type
                             (mime-type path)
                             "application/octet-stream"))
    (handle-if-modified-since time)
    (with-open-file (file path
                     :direction :input
                     :element-type '(unsigned-byte 8)
                     :if-does-not-exist nil)
      (setf (header-out "Last-Modified") (rfc-1123-date time)
            (content-length) (file-length file))
      (let ((out (send-headers)))
        (loop with buf = (make-array +buffer-length+ :element-type '(unsigned-byte 8))
              for pos = (read-sequence buf file)
              until (zerop pos)
              do (write-sequence buf out :end pos)
                 (finish-output out))))))

(defun create-static-file-dispatcher-and-handler (uri path &optional content-type)
  "Creates and returns a dispatch function which will dispatch to a
handler function which emits the file denoted by the pathname
designator PATH with content type CONTENT-TYPE if the SCRIPT-NAME of
the request matches the string URI.  If CONTENT-TYPE is NIL tries to
determine the content type via the file's suffix."
  ;; the dispatcher
  (lambda (request)
    (when (equal (script-name request) uri)
      ;; the handler
      (lambda ()
        (handle-static-file path content-type)))))

;; should pull in cl-fad for one function?
(defun directory-pathname-p (pathspec)
  "Returns NIL if PATHSPEC \(a pathname designator) does not designate
a directory, PATHSPEC otherwise.  It is irrelevant whether file or
directory designated by PATHSPEC does actually exist."
  (flet ((component-present-p (value)
           (and value (not (eql value :unspecific)))))
    (and 
     (not (component-present-p (pathname-name pathspec)))
     (not (component-present-p (pathname-type pathspec)))
     pathspec)))

(defun file-size (file)
  "Returns filesize in bytes, or NIL if it is a directory."
  (cl:ignore-errors
    (with-open-file (in file :direction :input)
      (file-length in))))

(defstruct file-details name date size (desc "" :type string))

(defun file-date-string (file-details)
  "Returns a descriptive string like \"15-Feb-2007\"."
  (multiple-value-bind
        (second minute hour date month year)
      (decode-universal-time (file-details-date file-details))
    (declare (ignore second minute hour))
    (let ((month-names
           '("Jan" "Feb" "Mar" "Apr" "May" "Jun"
             "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")))  
      (format nil "~2,'0d-~A-~D"
              date (nth (1- month) month-names) year))))

(defun file-size-string (file-details)
  "Returns a descriptive string like \"1.3K\" or \"8M\"."
  (let ((size (file-details-size file-details)))
    (cond ((not (numberp size)) "-&nbsp;")
          ((< size 1024) (format nil "~D&nbsp;" size))
          ((< size (* 1024 1024)) (format nil "~,1FK" (/ size 1024)))
          (t (format nil "~,1FM" (/ size (* 1024 1024)))))))

(defun maybe-integer-lessp (a b)
  "If both a & b are numbers, then it's just like #'<
Otherwise, non number goes first."
  (cond ((and (numberp a) (numberp b))
         (< a b))
        ((numberp b) t)
        (t nil)))

(defun sort-file-details (list &optional (sort-column "N") (sort-order "A"))
  "Sort file-details based on sort column and order."
  (let* ((sort-column (intern sort-column :keyword))
         (compare-func
          (ecase sort-column
            ((:M :S) #'maybe-integer-lessp)
            ((:N :D) #'string-lessp)))
         (selector
          (ecase sort-column
            (:N #'file-details-name)
            (:M #'file-details-date)
            (:S #'file-details-size)
            (:D #'file-details-desc)))
         (results (sort list compare-func :key selector)))
    (if (string-equal sort-order "D")
        (nreverse results)
        results)))

(defun redirect-if-malformed-pathspec ()
  "Directory listing request should end in a forward slash like
\"http://localhost/hunchentoot/code/\". Fix it if it is not the
case."
  (let* ((script-name (script-name))
         (length (length script-name)))
    (unless (and (> length 0)
                 (char= (aref script-name (1- length)) #\/))
      ;; stripping the query string is OK
      (redirect (concatenate 'string script-name "/")))))

(defun folder-index-page (pathname)
  "Returns a html page with a directory listing like those generated
by Apache."
  (redirect-if-malformed-pathspec)
  (let* ((contents (directory (namestring pathname)))
         (title (format nil "Index of ~A" (script-name)))
         (sort-column (get-parameter "C"))
         (sort-order (get-parameter "O")))
    (flet ((file-details (file)
             (make-file-details :name (enough-namestring file pathname)
                                :date (file-write-date file)
                                :size (file-size file))))
      ;; taint get-parameters
      (unless (member sort-column '("N" "M" "S" "D") :test #'string-equal)
        (setq sort-column "N"))
      (unless (member sort-order '("A" "D") :test #'string-equal)
        (setq sort-order "A"))
      ;; page-out
      (with-output-to-string (out)
        (format out "<html><head><title>~A</title>
	<style type=\"text/css\">
	<!--
        .size { text-align: right; }
	th { padding: 0.2em; text-align: left; }
	td { padding: 0.2em; }
	table { font-family: monospace;	}
	-->
	</style></head><body><h1>~A</h1><table><tr>" title title)
        ;; column headers
        (loop for (query-char desc) in
              '(("N" "Name")
                ("M" "Last modified")
                ("S" "Size")
                ("D" "Description")) do
                (format out "<th><a href=\"?C=~A&O=~A\">~A</a></th>"
                        query-char
                        (or (when (string-equal query-char sort-column)
                              (if (string-equal sort-order "D") "A" "D"))
                            "A")
                        desc))
        (format out "</tr><tr><td colspan=\"4\"><hr/></td></tr><tr>
                     <td><a href=\"../\">Parent Directory</a></td>
                     <td></td><td></td><td></td></tr>")
        (dolist (d (sort-file-details
                    (mapcar #'file-details contents)
                    sort-column sort-order))
          (format out "<tr><td><a href=\"~A\">~A</a></td>"
                  (file-details-name d) (file-details-name d))
          (format out "<td>~A</td><td class=\"size\">~A</td><td>~A</td></tr>"
                  (file-date-string d)
                  (file-size-string d) 
                  (file-details-desc d)))
        (format out "<tr><td colspan=\"4\"><hr/></td></tr></table>
                     <p>~A</p></body></html>" (address-string))))))

(defun create-folder-dispatcher-and-handler (uri-prefix base-path &optional content-type)
  "Creates and returns a dispatch function which will dispatch to a
handler function which emits the file relative to BASE-PATH that is
denoted by the URI of the request relative to URI-PREFIX.  URI-PREFIX
must be a string ending with a slash, BASE-PATH must be a pathname
designator for an existing directory.  If CONTENT-TYPE is not NIL,
it'll be the content type used for all files in the folder."
  (unless (and (stringp uri-prefix)
               (plusp (length uri-prefix))
               (char= (char uri-prefix (1- (length uri-prefix))) #\/))
    (error "~S must be string ending with a slash." uri-prefix))
  (when (or (pathname-name base-path)
            (pathname-type base-path))
    (error "~S is supposed to denote a directory." base-path))
  (flet ((handler ()
           (let* ((script-path (enough-namestring (regex-replace-all "\\\\" (script-name) "/")
                                                  uri-prefix))
                  (script-path-directory (pathname-directory script-path)))
             (unless (or (stringp script-path-directory)
                         (null script-path-directory)
                         (and (listp script-path-directory)
                              (eq (first script-path-directory) :relative)
                              (loop for component in (rest script-path-directory)
                                    always (stringp component))))
               (setf (return-code) +http-forbidden+)
               (throw 'handler-done nil))
             (let ((pathname (probe-file (merge-pathnames script-path base-path))))
               (if (and pathname (directory-pathname-p pathname))
                   (folder-index-page pathname)
                   (handle-static-file pathname content-type))))))
    (create-prefix-dispatcher uri-prefix #'handler)))

(defun no-cache ()
  "Adds appropriate headers to completely prevent caching on most browsers."
  (setf (header-out "Expires")
          "Mon, 26 Jul 1997 05:00:00 GMT"
        (header-out "Cache-Control")
          "no-store, no-cache, must-revalidate, post-check=0, pre-check=0"
        (header-out "Pragma")
          "no-cache"
        (header-out "Last-Modified")
          (rfc-1123-date))
  (values))

(defun ssl-p ()
  "Whether the current connection to the client is secure."
  (cond ((server-mod-lisp-p *server*) (ssl-session-id *request*))
        (t (server-ssl-certificate-file *server*))))

(defun redirect (target &key (host (host *request*) host-provided-p)
                             (protocol (if (ssl-p) :https :http))
                             (add-session-id (not (or host-provided-p
                                                      (starts-with-scheme-p target)
                                                      (cookie-in *session-cookie-name*))))
                             permanently)
  "Redirects the browser to TARGET which should be a string.
If TARGET is a full URL starting with a scheme, HOST and PROTOCOL
are ignored.  Otherwise, TARGET should denote the path part of a
URL, PROTOCOL must be one of the keywords :HTTP or :HTTPS, and
the URL to redirect to will be constructed from HOST, PROTOCOL,
and TARGET.  Adds a session ID if ADD-SESSION-ID is true.  If
PERMANENTLY is true, a 301 request is sent to the browser,
otherwise a 302."
  (let ((url (if (starts-with-scheme-p target)
               target
               (format nil "~A://~A~A"
                       (ecase protocol
                         ((:http) "http")
                         ((:https) "https"))
                       host target))))
    (when add-session-id
      (setq url (add-cookie-value-to-url url :replace-ampersands-p nil)))
    (setf (header-out "Location")
            url
          (return-code *reply*)
            (if permanently
              +http-moved-permanently+
              +http-moved-temporarily+))
    (throw 'handler-done nil)))

(defun require-authorization (&optional (realm "Hunchentoot"))
  "Sends back appropriate headers to require basic HTTP authentication
(see RFC 2617) for the realm REALM."
  (setf (header-out "WWW-Authenticate")
          (format nil "Basic realm=\"~A\"" (quote-string realm))
        (return-code *reply*)
          +http-authorization-required+)
  (throw 'handler-done nil))
